/******************************************************************************/
/*! @file Keys.cc
    @brief Keys class
    @author Tachibanamasashi, Apolloron Project.
 ******************************************************************************/

#include <stdio.h>
#include <string.h>
#include "apolloron.h"

namespace apolloron {


// change 1 byte value to 6bit vale (0-63)
#define b64(a) (((unsigned char)(a) & 0xfc) >> 2)

/*! Constructor of Keys.
    @param void
    @return void
 */
Keys::Keys() {
  int i;

  (*this).pElement = new Element ** [64];
  for (i = 0; i < 64; i++) {
    (*this).nElementTotal[i] = 0;
    (*this).nElementStart[i] = -1;
    (*this).nElementCache[i] = -1;
    (*this).pElement[i] = (Element **)NULL;
  }
}


/*! Constructor of Keys.
    @param keys: Keys for initialization.
    @return void
 */
Keys::Keys(const Keys &keys) {
  int i;

  (*this).pElement = new Element ** [64];
  for (i = 0; i < 64; i++) {
    (*this).nElementTotal[i] = 0;
    (*this).nElementStart[i] = -1;
    (*this).nElementCache[i] = -1;
    (*this).pElement[i] = (Element **)NULL;
  }

  set(keys);
}


/*! Destructor of Keys.
    @param void
    @return void
 */
Keys::~Keys() {
  (void)clear();
}

/*! Delete instance of Keys.
    @param void
    @retval true   success
    @retval false  failure
 */
bool Keys::clear() {
  int i;
  long l;

  if (pElement != NULL && *pElement != NULL) {
    for (i = 0; i < 64; i++) {
      l = 0;
      while (l < (*this).nElementTotal[i]) {
        if ((*this).pElement[i][l] != NULL) {
          if ((*this).pElement[i][l]->key != NULL) {
            delete [] (*this).pElement[i][l]->key;
          }
          delete (*this).pElement[i][l]->value;
          delete [] (*this).pElement[i][l];
        }
        l++;
      }
      delete *((*this).pElement[i]);
      (*this).nElementTotal[i] = 0;
    }
    delete [] *((*this).pElement);
    (*this).pElement = NULL;
  }

  return true;
}


/*! Get an element that have match key.
    @param key  key of object
    @return String object
 */
String& Keys::operator [] (const String &key) {
  return (*this).getValue(key.c_str());
}


/*! Get an element that have match key.
    @param key  key of object
    @return String object
 */
String& Keys::operator [] (const char *key) {
  return (*this).getValue(key);
}


/*! Get an element that have match key.
    @param key  key of object
    @return Keys object
 */
String& Keys::getValue(const char *key) {
  Element *element;

  element = NULL;

  if (key != NULL) {
    bool found;
    int start;
    found = false;
    start = b64(key[0]);
    if (0 <= (*this).nElementCache[start]) {
      element = (*this).pElement[start][(*this).nElementStart[start]];
    } else {
      do {
        if (0 < (*this).nElementTotal[start]) {
          element = (*this).pElement[start][(*this).nElementStart[start]];
          while (element != NULL && element->key != NULL) {
            int cmp;
            cmp = strcmp(key, element->key);
            if (cmp == 0) {
              // found!
              found = true;
              break;
            } else if (cmp < 0) {
              // small
              if (element->small == -1) break;
              element = (*this).pElement[start][element->small];
              (*this).nElementCache[start] = element->small;
            } else {
              // big
              if (element->big == -1) break;
              element = (*this).pElement[start][element->big];
              (*this).nElementCache[start] = element->big;
            }
          }
        }
      } while (found == false && (*this).addKey(key) == true);
    }
  }

  return *((*element).value);
}


/*! Check the existance of key.
    @param key  key of object
    @retval true  key exist
    @retval false key exist
 */
bool Keys::isKey(const String &key) const {
  return isKey(key.c_str());
}


/*! Check the existance of key.
    @param key  key of object
    @retval true  key exist
    @retval false key exist
 */
bool Keys::isKey(const char *key) const {
  bool found;
  Element *element;

  found = false;

  if (key != NULL) {
    int start;
    start = b64(key[0]);
    if (0 < (*this).nElementTotal[start]) {
      if (0 <= (*this).nElementCache[start]) {
        element = (*this).pElement[start][(*this).nElementStart[start]];
        if (!strcmp(key, element->key)) {
          found = true;
        }
      } else {
        element = (*this).pElement[start][(*this).nElementStart[start]];
        while (element != NULL && element->key != NULL) {
          int cmp;
          cmp = strcmp(key, element->key);
          if (cmp == 0) {
            // found!
            found = true;
            break;
          } else if (cmp < 0) {
            // small
            if (element->small == -1) break;
            element = (*this).pElement[start][element->small];
          } else {
            // big
            if (element->big == -1) break;
            element = (*this).pElement[start][element->big];
          }
        }
      }
    }
  }

  return found;
}


/*! get the Number of keys in array.
    @param void
    @return number of keys
 */
long Keys::max() const {
  int i;
  long total;

  total = 0;

  for (i = 0; i < 64; i++) {
    total += (*this).nElementTotal[i];
  }

  return total;
}


/*! Get key of array.
    @param index  index of key
    @return key string
 */
const char *Keys::key(long index) const {
  Element *element;
  element = array(index);
  return element->key;
}


/*! Setup of keys.
    @param keys  Keys object to set
    @retval true  success
    @retval false failure
 */
bool Keys::set(const Keys & keys) {
  long i;

  clear();

  i = 0;
  while (i < keys.max()) {
    addKey(keys.array(i)->key, *(keys.array(i)->value));
    i++;
  }

  return true;
}


/*! Setup of keys.
    @param keys  Keys object to set
    @return Keys object
 */
const Keys& Keys::operator = (const Keys &keys) {
  set(keys);

  return *this;
}


/*! Get value of key in array.
    @param keys  Keys object to set
    @return Keys object
 */
String& Keys::value(long index) const {
  Element *element;
  element = array(index);
  return *(element->value);
}


/*! Get element of array
    @param index index of Keys
    @return Element object
 */
Element* Keys::array(long index) const {
  Element *element;

  element = (Element *)NULL;

  if (0 <= index) {
    int i;
    int start;
    long total;

    start = -1;
    total = 0;
    for (i = 0; i < 64; i++) {
      total += (*this).nElementTotal[i];
      if (index < total) {
        start = i;
        break;
      }
    }

    if (0 <= start) {
      long internal_index;
      long *stack;
      long stacknum;
      enum {UP, DOWN} dest;
      long count;

      internal_index = index - (total - (*this).nElementTotal[start]);
      stack = new long[(*this).nElementTotal[start]];
      for (stacknum = 0; stacknum < (*this).nElementTotal[start]; stacknum++) {
        stack[stacknum] = -1;
      }
      stacknum = 0;
      stack[stacknum] = (*this).nElementStart[start];
      stacknum++;
      dest = DOWN;

      if (internal_index <= (*this).nElementTotal[start]) {
        count = 0;
        while (count < internal_index) {
          if (dest == DOWN) {
            // stack down
            if (0 <= (*this).pElement[start][stack[stacknum - 1]]->small) {
              stack[stacknum] = (*this).pElement[start][stack[stacknum - 1]]->small;
              stacknum++;
            } else {
              if (1 < stacknum) {
                count++;
                if (internal_index <= count) {
                  break;
                }
              }
              if (0 <= (*this).pElement[start][stack[stacknum - 1]]->big) {
                stack[stacknum] = (*this).pElement[start][stack[stacknum - 1]]->big;
                stacknum++;
              } else {
                stacknum--;
                dest = UP;
              }
            }
          } else {
            // stack up
            while (0 < stacknum) {
              long  big;
              big = (*this).pElement[start][stack[stacknum - 1]]->big;
              if (!(0 <= big && big == stack[stacknum])) {
                count++;
                if (internal_index <= count) {
                  break;
                }
              }
              if (0 <= big && big != stack[stacknum]) {
                stack[stacknum] = big;
                stacknum++;
                dest = DOWN;
                break;
              }
              stacknum--;
            }
          }
        }
      } else {
        count = (*this).nElementTotal[start] - 1;
        while (internal_index <= count) {
          if (dest == DOWN) {
            // stack down
            if (0 <= (*this).pElement[start][stack[stacknum - 1]]->big) {
              stack[stacknum] = (*this).pElement[start][stack[stacknum - 1]]->big;
              stacknum++;
            } else {
              if (1 < stacknum) {
                count--;
                if (count < internal_index) {
                  break;
                }
              }
              if (0 <= (*this).pElement[start][stack[stacknum - 1]]->small) {
                stack[stacknum] = (*this).pElement[start][stack[stacknum - 1]]->small;
                stacknum++;
              } else {
                stacknum--;
                dest = UP;
              }
            }
          } else {
            // stack up
            while (0 < stacknum) {
              long  small;
              small = (*this).pElement[start][stack[stacknum - 1]]->small;
              if (!(0 <= small && small == stack[stacknum])) {
                count--;
                if (count < internal_index) {
                  break;
                }
              }
              if (0 <= small && small != stack[stacknum]) {
                stack[stacknum] = small;
                stacknum++;
                dest = DOWN;
                break;
              }
              stacknum--;
            }
          }
        }
      }

      element = (*this).pElement[start][stack[stacknum - 1]];
      delete [] stack;
    }
  }

  return element;
}


/*! Append a key to array
    @param key   key of array
    @param value value to append
    @retval true  success
    @retval false failure
 */
bool Keys::addKey(const char *key, const String &value) {
  return addKey(key, value.c_str());
}


/*! Append a key to array
    @param key   key of array
    @param value value to append
    @retval true  success
    @retval false failure
 */
bool Keys::addKey(const char *key, const char *value) {
  if (key != NULL) {
    int start;

    start = b64(key[0]);
    if ((*this).nElementTotal[start] == 0) {
      (*this).nElementStart[start] = 0;
      (*this).pElement[start] = new (Element *);
      (*this).pElement[start][0] = new Element;
      (*this).pElement[start][0]->key = new char[strlen(key) + 1];
      strcpy((*this).pElement[start][0]->key, key);
      (*this).pElement[start][0]->value = new String;
      (String &)(*((*this).pElement[start][0]->value)) = value;
      (*this).pElement[start][0]->small = -1;
      (*this).pElement[start][0]->big = -1;
      (*this).nElementTotal[start] = 1;
    } else {
      Element *parent_element;
      Element *target_element;
      long parent;
      long target;

      parent = (*this).nElementStart[start];
      parent_element = (*this).pElement[start][parent];
      target = -1;
      while (parent_element != NULL && parent_element->key != NULL) {
        int cmp;
        cmp = strcmp(key, parent_element->key);
        if (cmp == 0) {
          // ** replace (same key found) **
          if (value != NULL) {
            delete parent_element->value;
            parent_element->value = new String;
            (String &)(*(parent_element->value)) = value;
          }
          break;
        } else {
          Element **element_p;
          if (cmp < 0) {
            // small
            target = parent_element->small;
          } else {
            // big
            target = parent_element->big;
          }
          if (target < 0) {
            // ** add **
            element_p = new Element * [(*this).nElementTotal[start] + 1];
            memcpy(element_p, (*this).pElement[start], sizeof(Element *) * (*this).nElementTotal[start]);
            delete (*this).pElement[start];
            (*this).pElement[start] = element_p;
            (*this).pElement[start][(*this).nElementTotal[start]] = new Element;
            (*this).pElement[start][(*this).nElementTotal[start]]->key = new char[strlen(key) + 1];
            strcpy((*this).pElement[start][(*this).nElementTotal[start]]->key, key);
            (*this).pElement[start][(*this).nElementTotal[start]]->value = new String;
            *((*this).pElement[start][(*this).nElementTotal[start]]->value) = value;
            (*this).pElement[start][(*this).nElementTotal[start]]->small = -1;
            (*this).pElement[start][(*this).nElementTotal[start]]->big = -1;
            if (cmp < 0) {
              parent_element->small = (*this).nElementTotal[start];
            } else {
              parent_element->big = (*this).nElementTotal[start];
            }
            (*this).nElementTotal[start]++;
            break;
          } else {
            char *small_key, *big_key;
            target_element = (*this).pElement[start][target];
            small_key = (0 <= target_element->small)?(*this).pElement[start][target_element->small]->key:NULL;
            big_key = (0 <= target_element->big)?(*this).pElement[start][target_element->big]->key:NULL;
            if (small_key != NULL && 0 < strcmp(small_key, key) && big_key != NULL && strcmp(key, big_key) < 0) {
              // insert
              element_p = new Element * [(*this).nElementTotal[start] + 1];
              memcpy(element_p, (*this).pElement[start], sizeof(Element *) * (*this).nElementTotal[start]);
              delete (*this).pElement[start];
              (*this).pElement[start] = element_p;
              (*this).pElement[start][(*this).nElementTotal[start]] = new Element;
              (*this).pElement[start][(*this).nElementTotal[start]]->key = new char[strlen(key) + 1];
              strcpy((*this).pElement[start][(*this).nElementTotal[start]]->key, key);
              (*this).pElement[start][(*this).nElementTotal[start]]->value = new String;
              (String &)(*((*this).pElement[start][(*this).nElementTotal[start]]->value)) = value;
              (*this).pElement[start][(*this).nElementTotal[start]]->small = -1;
              (*this).pElement[start][(*this).nElementTotal[start]]->big = -1;
              // process for cycling tree structure
              if (0 < strcmp(small_key, key) && strcmp(key, target_element->key) < 0) {
                if (cmp < 0) {
                  (*this).pElement[start][(*this).nElementTotal[start]]->big = parent_element->small;
                  (*this).pElement[start][(*this).nElementTotal[start]]->small = target_element->small;
                  target_element->small = -1;
                } else {
                  (*this).pElement[start][(*this).nElementTotal[start]]->big = parent_element->small;
                  (*this).pElement[start][(*this).nElementTotal[start]]->big = target_element->big;
                  target_element->big = -1;
                }
              } else {
                if (cmp < 0) {
                  (*this).pElement[start][(*this).nElementTotal[start]]->small = parent_element->big;
                  (*this).pElement[start][(*this).nElementTotal[start]]->small = target_element->small;
                  target_element->small = -1;
                } else {
                  (*this).pElement[start][(*this).nElementTotal[start]]->small = parent_element->big;
                  (*this).pElement[start][(*this).nElementTotal[start]]->big = target_element->big;
                  target_element->big = -1;
                }
              }
              if (cmp < 0) {
                parent_element->small = (*this).nElementTotal[start];
              } else {
                parent_element->big = (*this).nElementTotal[start];
              }
              (*this).nElementTotal[start]++;
              break;
            }
          }
        }
        parent = target;
        parent_element = (*this).pElement[start][parent];
      }
    }
  }

  return true;
}


/*! remove a key from array
    @param key   key of array
    @retval true  success
    @retval false failure
 */
bool Keys::delKey(const String &key) {
  return (*this).delKey(key.c_str());
}


/*! remove a key from array
    @param key   key of array
    @retval true  success
    @retval false failure
 */
bool Keys::delKey(const char *key) {
  if (key != NULL) {
    int start;

    start = b64(key[0]);
    if (0 < (*this).nElementTotal[start]) {
      Element *parent_element;
      Element *target_element;
      Element *child_element;
      long parent;
      long target;
      long child;

      parent = -1;
      parent_element = (Element *)NULL;
      target = (*this).nElementStart[start];
      target_element = (*this).pElement[start][target];
      while (target_element != NULL && target_element->key != NULL) {
        int cmp;
        cmp = strcmp(key, target_element->key);
        if (cmp == 0) {
          // ** remove (found) **
          if (0 <= target_element->small && 0 <= target_element->big) {
            long *child_p;
            if (parent < 0 || parent_element->small == target) {
              // biggest key of smaller node.
              child_p = &(target_element->small);
              child = target_element->small;
              child_element = (*this).pElement[start][child];
              while (0 <= child_element->big) {
                child_p = &(child_element->big);
                child = child_element->big;
                child_element = (*this).pElement[start][child];
              }
              *child_p = -1;
              child_element->big = target;
              if (parent < 0) {
                (*this).nElementStart[start] = child;
              } else {
                parent_element->small = child;
              }
            } else {
              // smallest ket of bigger node
              child_p = &(target_element->big);
              child = target_element->big;
              child_element = (*this).pElement[start][child];
              while (0 <= child_element->small) {
                child_p = &(child_element->small);
                child = child_element->small;
                child_element = (*this).pElement[start][child];
              }
              *child_p = -1;
              child_element->small = target;
              parent_element->big = child;
            }
          } else if (0 <= parent) {
            if (target_element->small < 0 && target_element->big < 0) {
              if (parent_element->small == target) {
                parent_element->small = -1;
              }
              if (parent_element->big == target) {
                parent_element->big = -1;
              }
            } else {
              child = (0 <= target_element->small)?target_element->small:target_element->big;
              if (parent_element->small == target) {
                parent_element->small = child;
              } else {
                parent_element->big = child;
              }
            }
          } else {
            child = (0 <= target_element->small)?target_element->small:target_element->big;
            (*this).nElementStart[start] = child;
          }
          if (target_element->key != NULL) {
            delete [] target_element->key;
          }
          delete target_element->value;
          delete target_element;
          (*this).nElementTotal[start]--;
          // remove element of array
          if (target < (*this).nElementTotal[start]) {
            int i;
            (*this).pElement[start][target] = (*this).pElement[start][(*this).nElementTotal[start]];
            i = 0;
            while (i <= (*this).nElementTotal[start]) {
              if ((*this).pElement[start][i]->small == (*this).nElementTotal[start]) {
                (*this).pElement[start][i]->small = target;
              }
              if ((*this).pElement[start][i]->big == (*this).nElementTotal[start]) {
                (*this).pElement[start][i]->big = target;
              }
              i++;
            }
            if ((*this).nElementStart[start] == (*this).nElementTotal[start]) {
              (*this).nElementStart[start] = target;
            }
          }
          break;
        }
        parent = target;
        parent_element = (*this).pElement[start][parent];
        if (cmp < 0) {
          // small
          target = target_element->small;
        } else {
          // big
          target = target_element->big;
        }
        if (target < 0) {
          // ** do nothing (key not found) **
          break;
        }
        target_element = (*this).pElement[start][target];
      }
    }
    (*this).nElementCache[start] = -1;
  }

  return true;
}


/*! Read configulation file into a key.
    @param filename  File name to read.
    @retval true  success
    @retval false failure
 */
bool Keys::loadConfigFile(const char *filename) {
  String str, line, key, value, tmp_str;
  long len, i, j, k;

  len = str.loadFile(filename);
  if (len < 0L) {
    return false;
  }

  for (i = 0L; i < len; i++) {
    for (j = 0L; i + j < len; j++) {
      if (str[i + j] == '\n') {
        break;
      }
    }
    if (0L < j) {
      line = str.mid(i, j).trim();
      if (0L < line.len() && line[0] != '#' && 0 < (k = line.searchChar('='))) {
        String p, p1;
        p = line;
        p1 = p.left(k).trimR();
        key = p1;
        p1 = p.mid(k+1).trimL();
        value = p1;
        if (2L <= value.len() && value[0] == '"' && value[value.len()-1] == '"') {
          tmp_str = value.mid(1, value.len()-1);
          value = tmp_str.unescapeQuote();
          tmp_str.clear();
        }
        (*this).addKey(key, value);
      }
      line.clear();
      i += j;
    }
  }

  str.clear();

  return true;
}

} // namespace apolloron
